@@ -3,13 +3,16 @@ package File::Slurp;
 use strict;
 use warnings ;
 
-our $VERSION = '9999.24';
+our $VERSION = '9999.31';
 $VERSION = eval $VERSION;
 
 use Carp ;
 use Exporter qw(import);
 use Fcntl qw( :DEFAULT ) ;
+use File::Basename ();
 use File::Spec;
+use File::Temp qw(tempfile);
+use IO::Handle ();
 use POSIX qw( :fcntl_h ) ;
 use Errno ;
 
@@ -59,188 +62,84 @@ my $is_win32 = $^O =~ /win32/i ;
 *rf = \&read_file ;
 
 sub read_file {
+	my $file_name = shift;
+	my $opts = (ref $_[0] eq 'HASH') ? shift : {@_};
+	# options we care about:
+	# array_ref binmode blk_size buf_ref chomp err_mode scalar_ref
 
-	my $file_name = shift ;
-	my $opts = ( ref $_[0] eq 'HASH' ) ? shift : { @_ } ;
-
-# this is the optimized read_file for shorter files.
-# the test for -s > 0 is to allow pseudo files to be read with the
-# regular loop since they return a size of 0.
-
-	if ( !ref $file_name && -e $file_name && -s _ > 0 &&
-	     -s _ < $max_fast_slurp_size && !%{$opts} && !wantarray ) {
-
-
-		my $fh ;
-		unless( sysopen( $fh, $file_name, O_RDONLY ) ) {
-
-			@_ = ( $opts, "read_file '$file_name' - sysopen: $!");
-			goto &_error ;
+	# let's see if we have a stringified object before doing anything else
+	# We then only have to deal with when we are given a file handle/globref
+	if (ref($file_name)) {
+		my $ref_result = _check_ref($file_name, $opts);
+		if (ref($ref_result)) {
+			@_ = ($opts, $ref_result);
+			goto &_error;
 		}
-
-		my $read_cnt = sysread( $fh, my $buf, -s _ ) ;
-
-		unless ( defined $read_cnt ) {
-
-			@_ = ( $opts,
-				"read_file '$file_name' - small sysread: $!");
-			goto &_error ;
-		}
-
-		$buf =~ s/\015\012/\n/g if $is_win32 ;
-		return $buf ;
+		$file_name = $ref_result if $ref_result;
+		# we have now stringified $file_name if possible. if it's still a ref
+		# then we probably have a file handle
 	}
 
-# set the buffer to either the passed in one or ours and init it to the null
-# string
-
-	my $buf ;
-	my $buf_ref = $opts->{'buf_ref'} || \$buf ;
-	${$buf_ref} = '' ;
-
-	my( $read_fh, $size_left, $blk_size ) ;
-
-# deal with ref for a file name
-# it could be an open handle or an overloaded object
-
-	if ( ref $file_name ) {
-
-		my $ref_result = _check_ref( $file_name ) ;
-
-		if ( ref $ref_result ) {
-
-# we got an error, deal with it
-
-			@_ = ( $opts, $ref_result ) ;
-			goto &_error ;
+	my $fh;
+	if (ref($file_name)) {
+		$fh = $file_name;
+	}
+	else {
+		# to keep with the old ways, read in :raw by default
+		unless (open $fh, "<:raw", $file_name) {
+			@_ = ($opts, "read_file '$file_name' - open: $!");
+			goto &_error;
 		}
-
-		if ( $ref_result ) {
-
-# we got an overloaded object and the result is the stringified value
-# use it as the file name
-
-			$file_name = $ref_result ;
-		}
-		else {
-
-# here we have just an open handle. set $read_fh so we don't do a sysopen
-
-			$read_fh = $file_name ;
-			$blk_size = $opts->{'blk_size'} || 1024 * 1024 ;
-			$size_left = $blk_size ;
+		# even though we set raw, let binmode take place here (busted)
+		if (my $bm = $opts->{binmode}) {
+			binmode $fh, $bm;
 		}
 	}
 
-# see if we have a path we need to open
-
-	unless ( $read_fh ) {
-
-# a regular file. set the sysopen mode
-
-		my $mode = O_RDONLY ;
-
-#printf "RD: BINARY %x MODE %x\n", O_BINARY, $mode ;
-
-		$read_fh = local( *FH ) ;
-#		$read_fh = gensym ;
-		unless ( sysopen( $read_fh, $file_name, $mode ) ) {
-			@_ = ( $opts, "read_file '$file_name' - sysopen: $!");
-			goto &_error ;
-		}
-
-		if ( my $binmode = $opts->{'binmode'} ) {
-			binmode( $read_fh, $binmode ) ;
-		}
-
-# get the size of the file for use in the read loop
-
-		$size_left = -s $read_fh ;
-
-#print "SIZE $size_left\n" ;
-
-# we need a blk_size if the size is 0 so we can handle pseudofiles like in
-# /proc. these show as 0 size but have data to be slurped.
-
-		unless( $size_left ) {
-
-			$blk_size = $opts->{'blk_size'} || 1024 * 1024 ;
-			$size_left = $blk_size ;
-		}
+	# we are now sure to have an open file handle. Let's slurp it in the same
+	# way that File::Slurper does.
+	my $buf;
+	my $buf_ref = $opts->{buf_ref} || \$buf;
+	${$buf_ref} = '';
+	my $blk_size = $opts->{blk_size} || 1024 * 1024;
+	if (my $size = -f $fh && -s _) {
+		$blk_size = $size if $size < $blk_size;
+		my ($pos, $read) = 0;
+		do {
+			unless(defined($read = read $fh, ${$buf_ref}, $blk_size, $pos)) {
+				@_ = ($opts, "read_file '$file_name' - read: $!");
+				goto &_error;
+			}
+			$pos += $read;
+		} while ($read && $pos < $size);
 	}
-
-# infinite read loop. we exit when we are done slurping
-
-	while( 1 ) {
-
-# do the read and see how much we got
-
-		my $read_cnt = sysread( $read_fh, ${$buf_ref},
-				$size_left, length ${$buf_ref} ) ;
-
-# since we're using sysread Perl won't automatically restart the call
-# when interrupted by a signal.
-
-		next if $!{EINTR};
-
-		unless ( defined $read_cnt ) {
-
-			@_ = ( $opts, "read_file '$file_name' - loop sysread: $!");
-			goto &_error ;
-		}
-
-# good read. see if we hit EOF (nothing left to read)
-
-		last if $read_cnt == 0 ;
-
-# loop if we are slurping a handle. we don't track $size_left then.
-
-		next if $blk_size ;
-
-# count down how much we read and loop if we have more to read.
-
-		$size_left -= $read_cnt ;
-		last if $size_left <= 0 ;
+	else {
+		${$buf_ref} = do { local $/; <$fh> };
 	}
+	seek($fh, $opts->{_data_tell}, SEEK_SET) if $opts->{_is_data} && $opts->{_data_tell};
 
-# fix up cr/lf to be a newline if this is a windows text file
-
-	${$buf_ref} =~ s/\015\012/\n/g if $is_win32 && !$opts->{'binmode'} ;
-
-	my $sep = $/ ;
-	$sep = '\n\n+' if defined $sep && $sep eq '' ;
-
-# see if caller wants lines
-
-	if( wantarray || $opts->{'array_ref'} ) {
-
-		use re 'taint' ;
+	# line endings if we're on Windows
+	${$buf_ref} =~ s/\015\012/\012/g if ${$buf_ref} && $is_win32 && !$opts->{binmode};
 
+	# we now have a buffer filled with the file content. Figure out how to
+	# return it to the user
+	my $want_array = wantarray; # let's only ask for this once
+	if ($want_array || $opts->{array_ref}) {
+		use re 'taint';
+		my $sep = $/;
+		$sep = '\n\n+' if defined $sep && $sep eq '';
+		# split the buffered content into lines
 		my @lines = length(${$buf_ref}) ?
-			${$buf_ref} =~ /(.*?$sep|.+)/sg : () ;
-
-		chomp @lines if $opts->{'chomp'} ;
-
-# caller wants an array ref
-
-		return \@lines if $opts->{'array_ref'} ;
-
-# caller wants list of lines
-
-		return @lines ;
+			${$buf_ref} =~ /(.*?$sep|.+)/sg : ();
+		chomp @lines if $opts->{chomp};
+		return \@lines if $opts->{array_ref};
+		return @lines;
 	}
-
-# caller wants a scalar ref to the slurped text
-
-	return $buf_ref if $opts->{'scalar_ref'} ;
-
-# caller wants a scalar with the slurped text (normal scalar context)
-
-	return ${$buf_ref} if defined wantarray ;
-
-# caller passed in an i/o buffer by reference (normal void context)
-
-	return ;
+	return $buf_ref if $opts->{scalar_ref};
+	# if the function was called in scalar context, return the contents
+	return ${$buf_ref} if defined $want_array;
+	# if we were called in void context, return nothing
+	return;
 }
 
 # errors in this sub are returned as scalar refs
@@ -249,7 +148,7 @@ sub read_file {
 
 sub _check_ref {
 
-	my( $handle ) = @_ ;
+	my( $handle, $opts ) = @_ ;
 
 # check if we are reading from a handle (GLOB or IO object)
 
@@ -257,7 +156,7 @@ sub _check_ref {
 
 # we have a handle. deal with seeking to it if it is DATA
 
-		my $err = _seek_data_handle( $handle ) ;
+		my $err = _seek_data_handle( $handle, $opts ) ;
 
 # return the error string if any
 
@@ -282,7 +181,10 @@ sub _check_ref {
 
 sub _seek_data_handle {
 
-	my( $handle ) = @_ ;
+	my( $handle, $opts ) = @_ ;
+	# store some meta-data about the __DATA__ file handle
+	$opts->{_is_data} = 0;
+	$opts->{_data_tell} = 0;
 
 # DEEP DARK MAGIC. this checks the UNTAINT IO flag of a
 # glob/handle. only the DATA handle is untainted (since it is from
@@ -304,11 +206,18 @@ ERR
 
 	if ( B::svref_2object( $handle )->IO->IoFLAGS & 16 ) {
 
+		# we now know we have the data handle. Let's store its original
+		# location in the file so that we can put it back after the read.
+		# this is only done for Bugwards-compatibility in some dists such as
+		# CPAN::Index::API that made use of the oddity where sysread was in use
+		# before
+		$opts->{_is_data} = 1;
+		$opts->{_data_tell} = tell($handle);
 # set the seek position to the current tell.
 
-		unless( sysseek( $handle, tell( $handle ), SEEK_SET ) ) {
-			return "read_file '$handle' - sysseek: $!" ;
-		}
+		# unless( sysseek( $handle, tell( $handle ), SEEK_SET ) ) {
+		# 	return "read_file '$handle' - sysseek: $!" ;
+		# }
 	}
 
 # seek was successful, return no error string
@@ -316,192 +225,118 @@ ERR
 	return ;
 }
 
-
 *wf = \&write_file ;
 
 sub write_file {
+	my $file_name = shift;
+	my $opts = (ref $_[0] eq 'HASH') ? shift : {};
+	# options we care about:
+	# append atomic binmode buf_ref err_mode no_clobber perms
 
-	my $file_name = shift ;
-
-# get the optional argument hash ref from @_ or an empty hash ref.
-
-	my $opts = ( ref $_[0] eq 'HASH' ) ? shift : {} ;
-
-	my( $buf_ref, $write_fh, $no_truncate, $orig_file_name, $data_is_ref ) ;
-
-# get the buffer ref - it depends on how the data is passed into write_file
-# after this if/else $buf_ref will have a scalar ref to the data.
-
-	if ( ref $opts->{'buf_ref'} eq 'SCALAR' ) {
-
-# a scalar ref passed in %opts has the data
-# note that the data was passed by ref
-
-		$buf_ref = $opts->{'buf_ref'} ;
-		$data_is_ref = 1 ;
-	}
-	elsif ( ref $_[0] eq 'SCALAR' ) {
-
-# the first value in @_ is the scalar ref to the data
-# note that the data was passed by ref
-
-		$buf_ref = shift ;
-		$data_is_ref = 1 ;
-	}
-	elsif ( ref $_[0] eq 'ARRAY' ) {
-
-# the first value in @_ is the array ref to the data so join it.
-
-		${$buf_ref} = join '', @{$_[0]} ;
-	}
-	else {
-
-# good old @_ has all the data so join it.
-
-		${$buf_ref} = join '', @_ ;
-	}
-
-# deal with ref for a file name
-
-	if ( ref $file_name ) {
-
-		my $ref_result = _check_ref( $file_name ) ;
-
-		if ( ref $ref_result ) {
-
-# we got an error, deal with it
-
-			@_ = ( $opts, $ref_result ) ;
-			goto &_error ;
+	my $fh;
+	my $no_truncate = 0;
+	my $orig_filename;
+	# let's see if we have a stringified object or some sort of handle
+	# or globref before doing anything else
+	if (ref($file_name)) {
+		my $ref_result = _check_ref($file_name, $opts);
+		if (ref($ref_result)) {
+			# some error happened while checking for a ref
+			@_ = ($opts, $ref_result);
+			goto &_error;
 		}
-
-		if ( $ref_result ) {
-
-# we got an overloaded object and the result is the stringified value
-# use it as the file name
-
-			$file_name = $ref_result ;
+		if ($ref_result) {
+			# we have now stringified $file_name from the overloaded obj
+			$file_name = $ref_result;
 		}
 		else {
-
-# we now have a proper handle ref.
-# make sure we don't call truncate on it.
-
-			$write_fh = $file_name ;
-			$no_truncate = 1 ;
+			# we now have a proper handle ref
+			# make sure we don't call truncate on it
+			$fh = $file_name;
+			$no_truncate = 1;
+			# can't do atomic or permissions on a file handle
+			delete $opts->{atomic};
+			delete $opts->{perms};
 		}
 	}
 
-# see if we have a path we need to open
-
-	unless( $write_fh ) {
-
-# spew to regular file.
-
-		if ( $opts->{'atomic'} ) {
-
-# in atomic mode, we spew to a temp file so make one and save the original
-# file name.
-			$orig_file_name = $file_name ;
-			$file_name .= ".$$" ;
+	# open the file for writing if we were given a filename
+	unless ($fh) {
+		$orig_filename = $file_name;
+		my $perms = defined($opts->{perms}) ? $opts->{perms} : 0666;
+		# set the mode for the sysopen
+		my $mode = O_WRONLY | O_CREAT;
+		$mode |= O_APPEND if $opts->{append};
+		$mode |= O_EXCL if $opts->{no_clobber};
+		if ($opts->{atomic}) {
+			# in an atomic write, we must open a new file in the same directory
+			# as the original to account for ACLs. We must also set the new file
+			# to the same permissions as the original unless overridden by the
+			# caller's request to set a specified permission set.
+			my $dir = File::Spec->rel2abs(File::Basename::dirname($file_name));
+			if (!defined($opts->{perms}) && -e $file_name && -f _) {
+				$perms = 07777 & (stat $file_name)[2];
+			}
+			# we must ensure we're using a good temporary filename (doesn't already
+			# exist). This is slower, but safer.
+			{
+				local $^W = 0; # AYFKM
+				(undef, $file_name) = tempfile('.tempXXXXX', DIR => $dir, OPEN => 0);
+			}
 		}
-
-# set the mode for the sysopen
-
-		my $mode = O_WRONLY | O_CREAT ;
-		$mode |= O_APPEND if $opts->{'append'} ;
-		$mode |= O_EXCL if $opts->{'no_clobber'} ;
-
-		my $perms = $opts->{perms} ;
-		$perms = 0666 unless defined $perms ;
-
-#printf "WR: BINARY %x MODE %x\n", O_BINARY, $mode ;
-
-# open the file and handle any error.
-
-		$write_fh = local( *FH ) ;
-#		$write_fh = gensym ;
-		unless ( sysopen( $write_fh, $file_name, $mode, $perms ) ) {
-
-			@_ = ( $opts, "write_file '$file_name' - sysopen: $!");
-			goto &_error ;
+		$fh = local *FH;
+		unless (sysopen($fh, $file_name, $mode, $perms)) {
+			@_ = ($opts, "write_file '$file_name' - sysopen: $!");
+			goto &_error;
 		}
 	}
-
-	if ( my $binmode = $opts->{'binmode'} ) {
-		binmode( $write_fh, $binmode ) ;
+	# we now have an open file handle as well as data to write to that handle
+	if (my $binmode = $opts->{binmode}) {
+		binmode($fh, $binmode);
 	}
 
-	sysseek( $write_fh, 0, SEEK_END ) if $opts->{'append'} ;
-
-#print 'WR before data ', unpack( 'H*', ${$buf_ref}), "\n" ;
-
-# fix up newline to write cr/lf if this is a windows text file
-
-	if ( $is_win32 && !$opts->{'binmode'} ) {
-
-# copy the write data if it was passed by ref so we don't clobber the
-# caller's data
-		$buf_ref = \do{ my $copy = ${$buf_ref}; } if $data_is_ref ;
-		${$buf_ref} =~ s/\n/\015\012/g ;
+	# get the data to print to the file
+	# get the buffer ref - it depends on how the data is passed in
+	# after this if/else $buf_ref will have a scalar ref to the data
+	my $buf_ref;
+	my $data_is_ref = 0;
+	if (ref($opts->{buf_ref}) eq 'SCALAR') {
+		# a scalar ref passed in %opts has the data
+		# note that the data was passed by ref
+		$buf_ref = $opts->{buf_ref};
+		$data_is_ref = 1;
+	}
+	elsif (ref($_[0]) eq 'SCALAR') {
+		# the first value in @_ is the scalar ref to the data
+		# note that the data was passed by ref
+		$buf_ref = shift;
+		$data_is_ref = 1;
+	}
+	elsif (ref($_[0]) eq 'ARRAY') {
+		# the first value in @_ is the array ref to the data so join it.
+		${$buf_ref} = join '', @{$_[0]};
+	}
+	else {
+		# good old @_ has all the data so join it.
+		${$buf_ref} = join '', @_;
 	}
 
-#print 'after data ', unpack( 'H*', ${$buf_ref}), "\n" ;
+	# seek and print
+	seek($fh, 0, SEEK_END) if $opts->{append};
+	print {$fh} ${$buf_ref};
+	truncate($fh, tell($fh)) unless $no_truncate;
+	close($fh);
 
-# get the size of how much we are writing and init the offset into that buffer
-
-	my $size_left = length( ${$buf_ref} ) ;
-	my $offset = 0 ;
-
-# loop until we have no more data left to write
-
-	do {
-
-# do the write and track how much we just wrote
-
-		my $write_cnt = syswrite( $write_fh, ${$buf_ref},
-				$size_left, $offset ) ;
-
-# since we're using syswrite Perl won't automatically restart the call
-# when interrupted by a signal.
-
-		next if $!{EINTR};
-
-		unless ( defined $write_cnt ) {
-
-			@_ = ( $opts, "write_file '$file_name' - syswrite: $!");
-			goto &_error ;
-		}
-
-# track how much left to write and where to write from in the buffer
-
-		$size_left -= $write_cnt ;
-		$offset += $write_cnt ;
-
-	} while( $size_left > 0 ) ;
-
-# we truncate regular files in case we overwrite a long file with a shorter file
-# so seek to the current position to get it (same as tell()).
-
-	truncate( $write_fh,
-		  sysseek( $write_fh, 0, SEEK_CUR ) ) unless $no_truncate ;
-
-	close( $write_fh ) ;
-
-# handle the atomic mode - move the temp file to the original filename.
-
-	if ( $opts->{'atomic'} && !rename( $file_name, $orig_file_name ) ) {
-
-		@_ = ( $opts, "write_file '$file_name' - rename: $!" ) ;
-		goto &_error ;
+	if ($opts->{atomic} && !rename($file_name, $orig_filename)) {
+		@_ = ($opts, "write_file '$file_name' - rename: $!");
+		goto &_error;
 	}
 
-	return 1 ;
+	return 1;
 }
 
 # this is for backwards compatibility with the previous File::Slurp module.
 # write_file always overwrites an existing file
-
 *overwrite_file = \&write_file ;
 
 # the current write_file has an append mode so we use that. this
@@ -818,7 +653,7 @@ directory.
 =head2 WARNING - PENDING DOOM
 
 Although you technically I<can>, do NOT use this module to work on file handles,
-pipes, sockets, standard IO, pseudo-files, or the C<DATA> handle. These are
+pipes, sockets, standard IO, or the C<DATA> handle. These are
 features implemented long ago that just really shouldn't be abused here.
 
 Be warned: this activity will lead to inaccurate encoding/decoding of data.
@@ -836,7 +671,7 @@ L<File::Slurp> implements the following functions.
 
 =head2 append_file
 
-	use File::Spec qw(append_file write_file);
+	use File::Slurp qw(append_file write_file);
 	my $res = append_file('/path/file', "Some text");
 	# same as
 	my $res = write_file('/path/file', {append => 1}, "Some text");
@@ -926,7 +761,7 @@ function.
 
 =head2 overwrite_file
 
-	use File::Spec qw(overwrite_file);
+	use File::Slurp qw(overwrite_file);
 	my $res = overwrite_file('/path/file', "Some text");
 
 The C<overwrite_file> function is simply a synonym for the
@@ -937,7 +772,7 @@ L<File::Slurp/"write_file"> function.
 	use File::Slurp qw(prepend_file);
 	prepend_file('/path/file', $header);
 	prepend_file('/path/file', \@lines);
-	prepend_file('/path/file', { binmode => 'raw:'}, $bin_data);
+	prepend_file('/path/file', { binmode => ':raw'}, $bin_data);
 
 	# equivalent to:
 	use File::Slurp qw(read_file write_file);
@@ -963,7 +798,7 @@ a consistent file.
 
 =head2 read_dir
 
-	use File::Spec qw(read_dir);
+	use File::Slurp qw(read_dir);
 	my @files = read_dir('/path/to/dir');
 	# all files, even the dots
 	my @files = read_dir('/path/to/dir', keep_dot_dot => 1);
@@ -1112,7 +947,7 @@ copy of the file to return.
 
 =head2 rf
 
-	use File::Spec qw(rf);
+	use File::Slurp qw(rf);
 	my $text = rf('/path/file');
 
 The C<rf> function is simply a synonym for the L<File::Slurp/"read_file">
@@ -1120,7 +955,7 @@ function.
 
 =head2 slurp
 
-	use File::Spec qw(slurp);
+	use File::Slurp qw(slurp);
 	my $text = slurp('/path/file');
 
 The C<slurp> function is simply a synonym for the L<File::Slurp/"read_file">
@@ -1128,7 +963,7 @@ function.
 
 =head2 wf
 
-	use File::Spec qw(wf);
+	use File::Slurp qw(wf);
 	my $res = wf('/path/file', "Some text");
 
 
@@ -1188,11 +1023,10 @@ atomic
 
 The C<atomic> option is a boolean option, defaulted to false (C<0>). Setting
 this option to true (C<1>) will cause the file to be be written to in an
-atomic fashion. A temporary file name is created by appending the pid
-(C<$$>) to the file name argument and that file is spewed to. After the
-file is closed it is renamed to the original file name (and C<rename> is
-an atomic operation on most OSes). If the program using this were to
-crash in the middle of this, then the file with the pid suffix could
+atomic fashion. A temporary file name is created using L<File::Temp/"tempfile">.
+After the file is closed it is renamed to the original file name
+(and C<rename> is an atomic operation on most OSes). If the program using
+this were to crash in the middle of this, then the temporary file could
 be left behind.
 
 =item
@@ -1259,6 +1093,27 @@ You can get all subs in the module exported with
 
 	use File::Slurp qw(:all);
 
+=head1 SEE ALSO
+
+=over
+
+=item *
+
+L<File::Slurper> - Provides a straightforward set of functions for the most
+common tasks of reading/writing text and binary files.
+
+=item *
+
+L<Path::Tiny> - Lightweight and comprehensive file handling, including simple
+methods for reading, writing, and editing text and binary files.
+
+=item *
+
+L<Mojo::File> - Similar to Path::Tiny for the L<Mojo> toolkit, always works in
+bytes.
+
+=back
+
 =head1 AUTHOR
 
 Uri Guttman, <F<uri@stemsystems.com>>
